var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { getItems, replaceItems } from "feathers-hooks-common";
import { subject } from "@casl/ability";
import _pick from "lodash/pick";
import _isEmpty from "lodash/isEmpty";
import hasRestrictingFields from "../../utils/hasRestrictingFields";
import { shouldSkip, mergeArrays } from "feathers-utils";
import { getPersistedConfig, restore$select, getAbility, makeOptions } from "./authorize.hook.utils";
import getModelName from "../../utils/getModelName";
import { Forbidden } from "@feathersjs/errors";
const HOOKNAME = "authorize";
export default (options) => {
    return (context) => __awaiter(void 0, void 0, void 0, function* () {
        const $select = restore$select(context);
        if (shouldSkip(HOOKNAME, context) ||
            context.type !== "after" ||
            !context.params) {
            return context;
        }
        options = makeOptions(context.app, options);
        const modelName = getModelName(options.modelName, context);
        if (!modelName) {
            return context;
        }
        const skipCheckConditions = getPersistedConfig(context, "skipRestrictingRead.conditions");
        const skipCheckFields = getPersistedConfig(context, "skipRestrictingRead.fields");
        if (skipCheckConditions && skipCheckFields) {
            return context;
        }
        const { params } = context;
        params.ability = yield getAbility(context, options);
        if (!params.ability) {
            // Ignore internal or not authenticated requests
            return context;
        }
        const { ability } = params;
        const items = getItems(context);
        const availableFields = (!(options === null || options === void 0 ? void 0 : options.availableFields))
            ? undefined
            : (typeof options.availableFields === "function")
                ? options.availableFields(context)
                : options.availableFields;
        const hasRestrictingFieldsOptions = {
            availableFields: availableFields
        };
        const pickFieldsForItem = (item) => {
            const method = (Array.isArray(items)) ? "find" : "get";
            if (!skipCheckConditions && !ability.can(method, subject(modelName, item))) {
                return undefined;
            }
            let fields = hasRestrictingFields(ability, method, subject(modelName, item), hasRestrictingFieldsOptions);
            if (fields === true) {
                // full restriction
                return {};
            }
            else if (skipCheckFields || (!fields && !$select)) {
                // no restrictions
                return item;
            }
            else if (fields && $select) {
                fields = mergeArrays(fields, $select, "intersect");
            }
            else {
                fields = (fields) ? fields : $select;
            }
            return _pick(item, fields);
        };
        let result;
        if (Array.isArray(items)) {
            result = [];
            for (let i = 0, n = items.length; i < n; i++) {
                const item = pickFieldsForItem(items[i]);
                if (item) {
                    result.push(item);
                }
            }
        }
        else {
            result = pickFieldsForItem(items);
            if (context.method === "get" && _isEmpty(result)) {
                if (options.actionOnForbidden)
                    options.actionOnForbidden();
                throw new Forbidden(`You're not allowed to ${context.method} ${modelName}`);
            }
        }
        replaceItems(context, result);
        return context;
    });
};
