import _get from "lodash/get";
import _set from "lodash/set";
import _unset from "lodash/unset";
import { Forbidden } from "@feathersjs/errors";
import { isMulti } from "feathers-utils";
export const makeOptions = (app, options) => {
    options = options || {};
    return Object.assign({}, defaultOptions, getAppOptions(app), options);
};
const defaultOptions = {
    ability: undefined,
    actionOnForbidden: undefined,
    availableFields: (context) => {
        var _a, _b;
        const availableFields = (_b = (_a = context.service.options) === null || _a === void 0 ? void 0 : _a.casl) === null || _b === void 0 ? void 0 : _b.availableFields;
        if (!availableFields)
            return undefined;
        return (typeof availableFields === "function")
            ? availableFields(context)
            : availableFields;
    },
    checkMultiActions: false,
    checkAbilityForInternal: false,
    modelName: (context) => {
        return context.path;
    }
};
export const makeDefaultOptions = (options) => {
    return Object.assign({}, defaultOptions, options);
};
const getAppOptions = (app) => {
    const caslOptions = app === null || app === void 0 ? void 0 : app.get("casl");
    return (caslOptions && caslOptions.authorizeHook)
        ? caslOptions.authorizeHook
        : {};
};
export const getAbility = (context, options) => {
    var _a, _b;
    // if params.ability is set, return it over options.ability
    if ((_a = context === null || context === void 0 ? void 0 : context.params) === null || _a === void 0 ? void 0 : _a.ability) {
        if (typeof context.params.ability === "function") {
            const ability = context.params.ability(context);
            return Promise.resolve(ability);
        }
        else {
            return Promise.resolve(context.params.ability);
        }
    }
    if (!options.checkAbilityForInternal && !((_b = context.params) === null || _b === void 0 ? void 0 : _b.provider)) {
        return Promise.resolve(undefined);
    }
    if (options === null || options === void 0 ? void 0 : options.ability) {
        if (typeof options.ability === "function") {
            const ability = options.ability(context);
            return Promise.resolve(ability);
        }
        else {
            return Promise.resolve(options.ability);
        }
    }
    return Promise.resolve(undefined);
};
const move = (context, fromPath, toPath) => {
    const val = _get(context, fromPath);
    if (val !== undefined) {
        _unset(context, fromPath);
        _set(context, toPath, val);
    }
    return val;
};
export const throwUnlessCan = (ability, method, resource, modelName, actionOnForbidden) => {
    if (ability.cannot(method, resource)) {
        if (actionOnForbidden)
            actionOnForbidden();
        throw new Forbidden(`You are not allowed to ${method} ${modelName}`);
    }
};
export const checkMulti = (context, ability, modelName, options) => {
    const { method } = context;
    const currentIsMulti = isMulti(context);
    if (!currentIsMulti) {
        return true;
    }
    if ((method === "find" && ability.can(method, modelName)) ||
        (ability.can(`${method}-multi`, modelName))) {
        return true;
    }
    if (options === null || options === void 0 ? void 0 : options.actionOnForbidden)
        options.actionOnForbidden();
    throw new Forbidden(`You're not allowed to multi-${method} ${modelName}`);
};
export const hide$select = (context) => {
    return move(context, "params.query.$select", "params.casl.$select");
};
export const restore$select = (context) => {
    move(context, "params.casl.$select", "params.query.$select");
    return _get(context, "params.query.$select");
};
export const get$select = (context) => {
    return getPersistedConfig(context, "$select");
};
export const setPersistedConfig = (context, key, val) => {
    return _set(context, `params.casl.${key}`, val);
};
export const getPersistedConfig = (context, key) => {
    return _get(context, `params.casl.${key}`);
};
