import _isPlainObject from "lodash/isPlainObject";
import { Forbidden } from "@feathersjs/errors";
import { mergeQuery } from "feathers-utils";
const invertedMap = {
    "$gt": "$lte",
    "$gte": "$lt",
    "$lt": "$gte",
    "$lte": "$gt",
    "$in": "$nin",
    "$nin": "$in",
    "$ne": (prop) => {
        return prop["$ne"];
    }
};
const supportedOperators = Object.keys(invertedMap);
const invertedProp = (prop, name) => {
    const map = invertedMap[name];
    if (typeof map === "string") {
        return { [map]: prop[name] };
    }
    else if (typeof map === "function") {
        return map(prop);
    }
};
const convertRuleToQuery = (rule, options) => {
    const { conditions, inverted } = rule;
    if (!conditions) {
        if (inverted && (options === null || options === void 0 ? void 0 : options.actionOnForbidden)) {
            options.actionOnForbidden();
        }
        return {};
    }
    if (inverted) {
        const newConditions = {};
        for (const prop in conditions) {
            if (_isPlainObject(conditions[prop])) {
                const obj = conditions[prop];
                for (const name in obj) {
                    if (!supportedOperators.includes(name)) {
                        console.error(`CASL: not supported property: ${name}`);
                        continue;
                    }
                    newConditions[prop] = invertedProp(obj, name);
                }
            }
            else {
                newConditions[prop] = { $ne: conditions[prop] };
            }
        }
        return newConditions;
    }
    else {
        return conditions;
    }
};
const getConditionalQueryFor = (ability, method, subjectType, options) => {
    options = options || {};
    options.actionOnForbidden = options.actionOnForbidden || (() => {
        throw new Forbidden("You're not allowed to make this request");
    });
    const rules = ability.rulesFor(method, subjectType);
    let query = {};
    for (let i = 0; i < rules.length; i++) {
        const rule = rules[i];
        const currentQuery = convertRuleToQuery(rule, options);
        query = mergeQuery(query, currentQuery, {
            defaultHandle: "combine"
        });
    }
    return query;
};
export default getConditionalQueryFor;
